#!/usr/bin/ruby

# collect_diffs.rb expects to find this script in the same directory as it
#

# TODO: exemplify syntax for 'cvs admin -m' when log message is missing
# TODO: make max-line limit on diff output configurable
# TODO: put max size limit on whole email
# TODO: support non-html mail too (text/plain, multipart/alternative)

$maxSubjectLength = 200
$maxLinesPerDiff = 1000

def blah(text)
  if $debug
    puts("cvsspam.rb: #{text}")
  end
end

def min(a, b)
  a<b ? a : b
end


# the regexps given as keys must not use subexpressions '(...)'
class MultiSub
  def initialize(hash)
    @mash = Array.new
    expr = nil
    hash.each do |key,val|
      if expr == nil ; expr="(" else expr<<"|(" end
      expr << key << ")"
      @mash << val
    end
    @re = Regexp.new(expr)
  end

  def gsub!(text)
    text.gsub!(@re) { |match|
      idx = -1
      $~.to_a.each { |subexp|
        break unless idx==-1 || subexp==nil
        idx += 1
      }
      idx==-1 ? match : @mash[idx].call(match)
    }
  end
end

class LogReader
  def initialize(logIO)
    @io = logIO
    advance
  end

  def currentLineCode ; @line[1,1]  end


  class ConstrainedIO
    def initialize(reader)
      @reader = reader
      @linecode = reader.currentLineCode
    end

    def each
      return if @reader == nil
      while true
        yield @reader.currentLine
        break unless @reader.advance && currentValid?
      end
      @reader = nil
    end

    def gets
      return nil if @reader == nil
      line = @reader.currentLine
      return nil if line==nil || !currentValid?
      @reader.advance
      return line
    end

    def currentValid?
      @linecode == @reader.currentLineCode
    end
  end

  def getLines
    ConstrainedIO.new(self)
  end

  def eof ; @line==nil  end

  def advance
    @line = @io.gets
    return false if @line == nil
    unless @line[0,1] == "#"
      raise "#{$logfile}:#{@io.lineno} line did not begin with '#': #{@line}"
    end
    return true
  end

  def currentLine
    @line==nil ? nil : @line[3, @line.length-4]
  end
end


def htmlEncode(text)
  text.gsub(/./) do
    case $&
      when "&" then "&amp;"
      when "<" then "&lt;"
      when ">" then "&gt;"
      when "\"" then "&quot;"
      else $&
    end
  end
end

# actually, allows '/' to appear
def urlEncode(text)
  text.sub(/[^a-zA-Z0-9\-,.*_\/]/) do
    "%#{sprintf('%2X', $&[0])}"
  end
end

class FileEntry
  def initialize(path)
    @path = path
    @lineAdditions = @lineRemovals = 0
  end

  attr_accessor :path, :type, :lineAdditions, :lineRemovals, :isBinary, :isEmpty, :tag, :fromVer, :toVer

  def file
    @path =~ /(.*)\/(.*)/
    $2
  end

  def basedir
    @path =~ /(.*)\/(.*)/
    $1
  end

  def removal?
    @type == "R"
  end

  def addition?
    @type == "A"
  end

  def modification?
    @type == "M"
  end
end


class LineConsumer
  def handleLines(lines, emailIO)
    @emailIO = emailIO
    @lineCount = 0
    setup
    lines.each do |line|
      @lineCount += 1
      consume(line)
    end
    teardown
  end

  def setup
  end

  def teardown
  end

  def lineno
    @lineCount
  end

  def println(text)
    @emailIO.puts(text)
  end

  def print(text)
    @emailIO.print(text)
  end
end


mailSub = proc { |match| "<a href=\"mailto:#{match}\">#{match}</a>" }
urlSub = proc { |match| "<a href=\"#{match}\">#{match}</a>" }
bugSub = proc { |match|
  match =~ /([0-9]+)/
  "<a href=\"#{$bugzillaURL.sub(/%s/, $1)}\">#{match}</a>"
}
commentSubstitutions = {
		'(?:mailto:)?[\w\.\-\+\=]+\@[\w\-]+(?:\.[\w\-]+)+\b' => mailSub,
		'\b(?:http|https|ftp):[^ \t\n<>"]+[\w/]' => urlSub}


class CommentHandler < LineConsumer
  def setup
    @haveBlank = false
    @comment = ""
  end

  def consume(line)
    if line =~ /^\s*$/
      @haveBlank = true
    else
      if @haveBlank
        @comment += "\n"
        @haveBlank = false
      end
      $mailSubject = line if $mailSubject == nil
      @comment += line += "\n"
    end
  end

  def teardown
    unless @comment == @lastComment
      println("<pre class=\"comment\">")
      encoded = htmlEncode(@comment)
      $commentEncoder.gsub!(encoded)
      println(encoded)
      println("</pre>")
      @lastComment = @comment
    end
  end
end


class TagHandler < LineConsumer
  def consume(line)
    # TODO: check there is only one line
    @tag = line
  end

  def getLastTag
    tmp = @tag
    @tag = nil
    tmp
  end
end

class VersionHandler < LineConsumer
  def consume(line)
    # TODO: check there is only one line
    $fromVer,$toVer = line.split(/,/)
  end
end


class FileHandler < LineConsumer
  def setTagHandler(handler)
    @tagHandler = handler
  end

  def setup
    $fileHeaderHtml = ''
    println("<hr noshade=\"nodhade\" size=\"1\" /><a name=\"file#{$fileEntries.size+1}\" /><div class=\"file\">")
  end

  def consume(line)
    $file = FileEntry.new(line)
    if $subjectTag == nil
      line =~ /[^\/]+/
      $subjectTag = $&
    end
    $fileEntries << $file
    findCommonPath($file.basedir)
    $file.tag = getTag
    handleFile($file)
  end

 protected
  def getTag
    @tagHandler.getLastTag
  end

  def findCommonPath(path)
    path = path.dup
    if $commonPath == nil
      $commonPath = path
    else
      until $commonPath == path
        if $commonPath.size>path.size
	  $commonPath.sub!(/(.*)\/.*$/, '\1')
        else
	  path.       sub!(/(.*)\/.*$/, '\1')
        end
      end
    end
  end

  def println(text)
    $fileHeaderHtml << text << "\n"
  end

  def print(text)
    $fileHeaderHtml << text
  end
end

# if using viewcvs, create a link to the named path's dir listing
def viewcvsPath(path)
  if $viewcvsURL == nil
    htmlEncode(path)
  else
    "<a href=\"#{$viewcvsURL}#{urlEncode(path)}\">#{htmlEncode(path)}</a>"
  end
end

def viewcvsFile(path, version)
  if $viewcvsURL == nil
    version
  else
    "<a href=\"#{$viewcvsURL}#{urlEncode(path)}?rev=#{version}&content-type=text/vnd.viewcvs-markup\">#{version}</a>"
  end
end


# in need of refactoring...

class AddedFileHandler < FileHandler
  def handleFile(file)
    file.type="A"
    file.toVer=$toVer
    print("<span class=\"pathname\" id=\"added\">")
    print(viewcvsPath(file.basedir))
    println("<br /></span>")
    println("<div class=\"fileheader\" id=\"added\"><big><b>#{htmlEncode(file.file)}</b></big> <small id=\"info\">added at #{viewcvsFile(file.path,file.toVer)}</small></div>")
  end
end

class RemovedFileHandler < FileHandler
  def handleFile(file)
    file.type="R"
    file.fromVer=$fromVer
    print("<span class=\"pathname\" id=\"removed\">")
    print(viewcvsPath(file.basedir))
    println("<br /></span>")
    println("<div class=\"fileheader\" id=\"removed\"><big><b>#{htmlEncode(file.file)}</b></big> <small id=\"info\">removed after #{viewcvsFile(file.path,file.fromVer)}</small></div>")
  end
end

def viewcvsDiff(file)
  if $viewcvsURL == nil || file.isBinary
    "-&gt;"
  else
    "<a href=\"#{$viewcvsURL}#{urlEncode(file.path)}.diff?r1=#{file.fromVer}&r2=#{file.toVer}\">-&gt;</a>"
  end
end

class ModifiedFileHandler < FileHandler
  def handleFile(file)
    file.type="M"
    file.fromVer=$fromVer
    file.toVer=$toVer
    print("<span class=\"pathname\">")
    print(viewcvsPath(file.basedir))
    println("<br /></span>")
    println("<div class=\"fileheader\"><big><b>#{htmlEncode(file.file)}</b></big> <small id=\"info\">#{viewcvsFile(file.path,file.fromVer)} #{viewcvsDiff(file)} #{viewcvsFile(file.path,file.toVer)}</small></div>")
  end
end


class UnifiedDiffStats
  def initialize
    @diffLines=3  # the three initial lines in the unidiff
  end

  def diffLines
    @diffLines
  end

  def consume(line)
    @diffLines += 1
    case line[0,1]
      when "+" then $file.lineAdditions += 1
      when "-" then $file.lineRemovals += 1
    end
  end
end

# TODO: change-within-line colourisation should really be comparing the
#       set of lines just removed with the set of lines just added, but
#       it currently considers just a single line

class UnifiedDiffColouriser < LineConsumer
  def initialize
    @currentState = "@"
    @currentStyle = "info"
  end

  def output=(io)
    @emailIO = io
  end

  def consume(line)
    initial = line[0,1]
    if initial != @currentState
      prefixLen = 1
      suffixLen = 0
      if initial=="+" && @currentState=="-" && @lineJustDeleted!=nil
        # may be an edit, try to highlight the changes part of the line
        a = line[1,line.length-1]
        b = @lineJustDeleted[1,@lineJustDeleted.length-1]
        prefixLen = commonPrefixLength(a, b)+1
        suffixLen = commonPrefixLength(a.reverse, b.reverse)
	# prevent prefix/suffux having overlap,
	suffixLen = min(suffixLen, min(line.length,@lineJustDeleted.length)-prefixLen)
	deleteInfixSize = @lineJustDeleted.length - (prefixLen+suffixLen)
        if prefixLen==1 && suffixLen==0 || deleteInfixSize<=0 || deleteInfixSize*100/@lineJustDeleted.length>33
          println(htmlEncode(@lineJustDeleted))
        else
          print(htmlEncode(@lineJustDeleted[0,prefixLen]))
          print("<span id=\"removedchars\">")
          print(formatChange(@lineJustDeleted[prefixLen,deleteInfixSize]))
          print("</span>")
          println(htmlEncode(@lineJustDeleted[@lineJustDeleted.length-suffixLen,suffixLen]))
        end
	@lineJustDeleted = nil
      end
      if initial=="-"
        @lineJustDeleted=line
	shift(initial)
        # we'll print it next time (fingers crossed)
        return
      elsif @lineJustDeleted!=nil
        println(htmlEncode(@lineJustDeleted))
        @lineJustDeleted = nil
      end
      shift(initial)
      addInfixSize = line.length - (prefixLen+suffixLen)
      if prefixLen==1 && suffixLen==0 || addInfixSize<=0 || addInfixSize*100/line.length>33
        encoded = htmlEncode(line)
      else
        encoded = htmlEncode(line[0,prefixLen]) +
        "<span id=\"addedchars\">" +
        formatChange(line[prefixLen,addInfixSize]) +
        "</span>" +
        htmlEncode(line[line.length-suffixLen,suffixLen])
      end
    else
      encoded = htmlEncode(line)
    end
    if initial=="-"
      unless @lineJustDeleted==nil
        println(htmlEncode(@lineJustDeleted))
        @lineJustDeleted=nil
      end
    end
    if initial=="+"
      if line =~ /\b(TODO\b.*)/
        $todoList << $1
        encoded.sub!(/\bTODO\b/, "<span id=\"TODO\">TODO</span>")
        encoded = "<a name=\"todo#{$todoList.size}\" />#{encoded}"
      elsif line =~ /\b(FIXME\b.*)/
        $todoList << $1
        encoded.sub!(/\FIXME\b/, "<span id=\"TODO\">FIXME</span>")
        encoded = "<a name=\"todo#{$todoList.size}\" />#{encoded}"
      end
      # won't bother with /XXX/ (false positives more likely?)
    end
    println(encoded)
  end

  def teardown
    unless @lineJustDeleted==nil
      println(htmlEncode(@lineJustDeleted))
      @lineJustDeleted = nil
    end
    shift(nil)
  end

 private

  def formatChange(text)
    return '<small id="info">^M</small>' if text=="\r"
    htmlEncode(text).gsub(/ /, '&nbsp;')
  end

  def shift(nextState)
    unless @currentState == nil
      if @currentStyle == "info"
        print("</small></pre>")
      else
        print("</pre>")
      end
      @currentStyle = case nextState
        when "\\" then "info" # as in '\ No newline at end of file'
        when "@" then "info"
        when " " then "context"
        when "+" then "added"
        when "-" then "removed"
      end
      unless nextState == nil
        if @currentStyle=='info'
          print("<pre class=\"diff\"><small id=\"info\">")
        else
          print("<pre class=\"diff\" id=\"#{@currentStyle}\">")
        end
      end
    end
    @currentState = nextState
  end

  def commonPrefixLength(a, b)
    length = 0
    a.each_byte do |char|
      break unless b[length]==char
      length = length + 1
    end
    return length
  end
end


class UnifiedDiffHandler < LineConsumer
  def setup
    @stats = UnifiedDiffStats.new
    @colour = UnifiedDiffColouriser.new
    @colour.output = @emailIO
    @lookahead = nil
    # evil here:
    def self.consume(line)
      case lineno
        when 1
         @diffline = line
        when 2
          @lookahead = line
        when 3
	  println($fileHeaderHtml)
	  # TODO: move to UnifiedDiffColouriser
	  print("<pre class=\"diff\"><small id=\"info\">")
	  println(htmlEncode(@diffline))  # 'diff ...'
	  println(htmlEncode(@lookahead)) # '--- ...'
	  println(htmlEncode(line))      # '+++ ...'

	  # now, redefine this method for the subsequent lines of input
	  def self.consume(line)
	    @stats.consume(line)
	    if @stats.diffLines < $maxLinesPerDiff
	      @colour.consume(line)
	    elsif @stats.diffLines == $maxLinesPerDiff
	      @colour.consume(line)
	      @colour.teardown
	    end
	  end
      end
    end
  end

  def teardown
    if @lookahead == nil
      $file.isEmpty = true
    elsif @lookahead  =~ /Binary files .* and .* differ/
      $file.isBinary = true
    else
      if @stats.diffLines > $maxLinesPerDiff
        println("</pre>")
        println("<strong class=\"error\">[truncated at #{$maxLinesPerDiff} lines; #{@stats.diffLines-$maxLinesPerDiff} more skipped]</strong>")
      else
        @colour.teardown
      end
      println("</div>") # end of "file" div
    end
  end
end









$config = "#{ENV['CVSROOT']}/CVSROOT/cvsspam.conf"

$debug = false
$recipients = Array.new
$sendmail_prog = "/usr/sbin/sendmail"

require 'getoptlong'

opts = GetoptLong.new(
  [ "--to",     "-t", GetoptLong::REQUIRED_ARGUMENT ],
  [ "--config", "-c", GetoptLong::REQUIRED_ARGUMENT ],
  [ "--debug",  "-d", GetoptLong::NO_ARGUMENT ]
)

opts.each do |opt, arg|
  $recipients << arg if opt=="--to"
  $config = arg if opt=="--config"
  $debug = true if opt=="--debug"
end


if ARGV.length != 1
  if ARGV.length > 1
    $stderr.puts "extra arguments not needed: #{ARGV[1, ARGV.length-1].join(', ')}"
  else
    $stderr.puts "missing required file argument"
  end
  puts "Usage: cvsspam.rb [ --to <email> ] [ --config <file> ] <collect_diffs file>"
  exit(-1)
end

$logfile = ARGV[0]


$additionalHeaders = Array.new
$problemHeaders = Array.new

# helper functions called from the 'config file'
def addHeader(name, value)
  if name =~ /^[!-9;-~]+$/
    $additionalHeaders << [name, value]
  else
    $problemHeaders << [name, value]
  end
end
def addRecipient(email)
  $recipients << email
end

if FileTest.exists?($config)
  load $config
else
  blah("Config file '#{$config}' not found, ignoring")
end

if $recipients.empty?
  fail "No email recipients defined"
end

$viewcvsURL << "/" if $viewcvsURL!=nil && $viewcvsURL!~/\/$/


unless $bugzillaURL == nil
  commentSubstitutions['\b[Bb][Uu][Gg]\s*#?[0-9]+'] = bugSub
end
$commentEncoder = MultiSub.new(commentSubstitutions)


tagHandler = TagHandler.new

$handlers = Hash[">" => CommentHandler.new,
		 "U" => UnifiedDiffHandler.new,
		 "T" => tagHandler,
		 "A" => AddedFileHandler.new,
		 "R" => RemovedFileHandler.new,
		 "M" => ModifiedFileHandler.new,
		 "V" => VersionHandler.new]

$handlers["A"].setTagHandler(tagHandler)
$handlers["R"].setTagHandler(tagHandler)
$handlers["M"].setTagHandler(tagHandler)

$fileEntries = Array.new
$todoList = Array.new

File.open("#{$logfile}.emailtmp", File::RDWR|File::CREAT|File::TRUNC) do |mail|

  File.open($logfile) do |log|
    reader = LogReader.new(log)

    until reader.eof
      handler = $handlers[reader.currentLineCode]
      if handler == nil
        raise "No handler file lines marked '##{reader.currentLineCode}'"
      end
      handler.handleLines(reader.getLines, mail)
    end
  end
end

if $subjectPrefix == nil
  $subjectPrefix = "[CVS #{$subjectTag}]"
end
mailSubject = "#{$subjectPrefix} #{$mailSubject}"
if mailSubject.length > $maxSubjectLength
  mailSubject = mailSubject[0, $maxSubjectLength]
end

blah("invoking #{$sendmail_prog} -t")
IO.popen("#{$sendmail_prog} -t", "w") do |mail|
  mail.puts("To: #{$recipients.join(',')}")
  mail.puts("Subject: #{mailSubject}")
  mail.puts("Content-Type: text/html")
  unless ($additionalHeaders.empty?)
    $additionalHeaders.each do |header|
      mail.puts("#{header[0]}: #{header[1]}")
    end
  end
  mail.puts  # end-of-headers

  mail.puts(<<HEAD)
<html>
<head>
<style><!--
  .file {border-width:1px;border-style:solid;border-color:#eeeeee;margin-top:1em;margin-bottom:1em;}
  .pathname {font-family:monospace; float:right;}
  .fileheader {margin-bottom:.5em;}
  .diff {margin:0;}
  .todolist {padding:4px;border:1px;border-color:#000000;border-style:dashed;margin-top:1em;}
  .todolist ul {margin-top:0;margin-bottom:0;}
  tr.alt {background-color:#eeeeee}
  #added {background-color:#ddffdd;}
  #addedchars {background-color:#99ff99;font-weight:bolder;}
  tr.alt #added {background-color:#ccf7cc;}
  #removed {background-color:#ffdddd;}
  #removedchars {background-color:#ff9999;font-weight:bolder;}
  tr.alt #removed {background-color:#f7cccc;}
  #info {color:#888888;}
  #context {background-color:#eeeeee;}
  #todo {background-color:#ffff00;}
  .comment {padding:4px;border:1px;border-color:#000000;border-style:dashed;background-color:#ffffdd}
  .error {color:red;}
--></style>
</head>
<body>
HEAD

  unless ($problemHeaders.empty?)
    mail.puts("<strong class=\"error\">Bad header format in '#{ARGV[0]}':<ul>")
    $stderr.puts("Bad header format in '#{ARGV[0]}':")
    $problemHeaders.each do |header|
      mail.puts("<li><pre>#{htmlEncode(header[0])}</pre></li>")
      $stderr.puts(" - #{header[0]}")
    end
    mail.puts("</ul></strong>")
  end
  mail.puts("Commit in <b><tt>#{htmlEncode($commonPath)}</tt></b>")
  mail.puts("<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\">")
  filesAdded = 0
  filesRemoved = 0
  filesModified  = 0
  totalLinesAdded = 0
  totalLinesRemoved = 0
  count = 0
  lastPath = ""
  $fileEntries.each do |file|
    count += 1
    if (count%2==0)
      mail.print("<tr class=\"alt\">")
    else
      mail.print("<tr>")
    end
    if file.addition?
      filesAdded += 1
    elsif file.removal?
      filesRemoved += 1
    elsif file.modification?
      filesModified += 1
    end
    name = htmlEncode(file.path.slice($commonPath.size+1,file.path.size-$commonPath.size-1))
    slashPos = name.rindex("/")
    if slashPos==nil
      prefix = ""
    else
      thisPath = name[0,slashPos]
      name = name[slashPos+1,name.length]
      if thisPath == lastPath
        prefix = "&nbsp;"*(slashPos) + "/"
      else 
        prefix = thisPath + "/"
      end
      lastPath = thisPath
    end
    if file.addition?
      name = "<span id=\"added\">#{name}</span>"
    elsif file.removal?
      name = "<span id=\"removed\">#{name}</span>"
    end
    if file.isEmpty || file.isBinary
      mail.print("<td><tt>#{prefix}#{name}&nbsp;</tt></td>")
    else
      mail.print("<td><tt>#{prefix}<a href=\"#file#{count}\">#{name}</a>&nbsp;</tt></td>")
    end
    if file.isEmpty
      mail.print("<td colspan=\"2\" align=\"center\"><small id=\"info\">[empty]</small></td>")
    elsif file.isBinary
      mail.print("<td colspan=\"2\" align=\"center\"><small id=\"info\">[binary]</small></td>")
    else
      if file.lineAdditions>0
        totalLinesAdded += file.lineAdditions
        mail.print("<td align=\"right\" id=\"added\">+#{file.lineAdditions}</td>")
      else
	mail.print("<td>&nbsp;</td>")
      end
      if file.lineRemovals>0
        totalLinesRemoved += file.lineRemovals
        mail.print("<td align=\"right\" id=\"removed\">-#{file.lineRemovals}</td>")
      else
        mail.print("<td>&nbsp;</td>")
      end
    end
    if file.tag
      mail.print("<td>#{htmlEncode(file.tag)}</td>")
    else
      mail.print("<td>&nbsp;</td>")
    end
    mail.print("<td>&nbsp;</td>")
    if file.addition?
      mail.print("<td align=\"right\">added #{viewcvsFile(file.path,file.toVer)}</td>")
    elsif file.removal?
      mail.print("<td>#{viewcvsFile(file.path,file.fromVer)} removed</td>")
    elsif file.modification?
      mail.print("<td align=\"center\">#{viewcvsFile(file.path,file.fromVer)} #{viewcvsDiff(file)} #{viewcvsFile(file.path,file.toVer)}</td>")
    end

    mail.puts("</tr>")
  end
  if $fileEntries.size>1
    # give total number of lines added/removed accross all files
    mail.print("<tr><td>&nbsp;</td>")
    if totalLinesAdded>0
      mail.print("<td align=\"right\" id=\"added\">+#{totalLinesAdded}</td>")
    else
      mail.print("<td>&nbsp;</td>")
    end
    if totalLinesRemoved>0
      mail.print("<td align=\"right\" id=\"removed\">-#{totalLinesRemoved}</td>")
    else
      mail.print("<td>&nbsp;</td>")
    end
    mail.puts("<td>&nbsp;</td></tr>")
  end
  
  mail.puts("</table>")

  totalFilesChanged = filesAdded+filesRemoved+filesModified
  if totalFilesChanged > 1
    mail.print("<small id=\"info\">")
    changeKind = 0
    if filesAdded>0
      mail.print("#{filesAdded} added")
      changeKind += 1
    end
    if filesRemoved>0
      mail.print(" + ") if changeKind>0
      mail.print("#{filesRemoved} removed")
      changeKind += 1
    end
    if filesModified>0
      mail.print(" + ") if changeKind>0
      mail.print("#{filesModified} modified")
      changeKind += 1
    end
    mail.print(", total #{totalFilesChanged}")
    mail.puts(" files</small><br />")
  end

  count = 0
  $todoList.each do |item|
    if count == 0
      mail.puts("<div class=\"todolist\"><center><small id=\"info\">Possibly new TODO items</small></center>")
      mail.puts("<ul>")
    end
    count += 1
    item = htmlEncode(item)
    mail.puts("<li><a href=\"#todo#{count}\">#{item}</a></li>")
  end
  mail.puts("</ul></div>") if count>0


  File.open("#{$logfile}.emailtmp") do |input|
    input.each do |line|
      mail.puts(line)
    end
  end
  if $debug
    blah("leaving file #{$logfile}.emailtmp")
  else
    File.unlink("#{$logfile}.emailtmp")
  end

  mail.puts("</body></html>")

end
